#! /usr/bin/ruby
#
# Copyright(C) 2010  Yuichi Kobayashi <kobayasy@kobayasy.com>
# Last modified: Sun Jun  6 01:01:03 JST 2010 (kobayasy)
#
# This program comes with NO WARRANTY, to the extent permitted by law.
# You may redistribute it under the terms of the GNU General Public License;
# see the file named LICENSE.txt for details.
#
$KCODE = 'UTF8'

HeaderMessage = nil
FooterMessage = nil
CGIURL, ImageURL = 'http://knavi.kobayasy.com', '%02d.png'
GoogleMapsAPIKey = 'ABQIAAAACZoSuum48ql8yDHo-nMT5hQIFTjONbH5F5l_m4UFSq0cAWT39xQH9EaJ8J6s0hpvN7ZQAQ62G-J_bw'

TimeNow = Time.now
HTMLTitle, HelpURL = '携ナビ', 'http://kobayasy.com/knavi/'
BBSURL = nil
CookieName, CookieExpires = 'knavi', TimeNow + 60*60*24*100

include Math

class CalcDatum

  Parameter = {
#   測地系                      赤道半径     扁平率           座標原点シフト量(X,Y,Z)
    'WGS84'                 => [6378137.0,   1/298.257223563,    0.0,     0.0,     0.0 ],
    'Tokyo'                 => [6377397.155, 1/299.1528128,   -146.36, +506.85, +680.27],
    'Adindan'               => [6378249.145, 1/293.465,       -162.0,   -12.0,  +206.0 ],
    'Arc1950'               => [6378249.145, 1/293.465,       -143.0,   -90.0,  -294.0 ],
    'Arc1960'               => [6378249.145, 1/293.465,       -160.0,    -8.0,  -300.0 ],
    'Australian1966'        => [6378160.0,   1/298.25,        -133.0,   -48.0,  +148.0 ],
    'Australian1984'        => [6378160.0,   1/298.25,        -134.0,   -48.0,  +149.0 ],
    'Camp Area Astro'       => [6378388.0,   1/297.0,         -104.0,  -129.0,  +239.0 ],
    'Cape'                  => [6378249.145, 1/293.465,       -136.0,  -108.0,  -292.0 ],
    'European 1950'         => [6378388.0,   1/297.0,          -87.0,   -98.0,  -121.0 ],
    'European 1979'         => [6378388.0,   1/297.0,          -86.0,   -98.0,  -119.0 ],
    'Geodetic Datum 1949'   => [6378388.0,   1/297.0,          +84.0,   -22.0,  +209.0 ],
    'Hong Kong 1963'        => [6378388.0,   1/297.0,         -156.0,  -271.0,  -189.0 ],
    'Hu-Tzu-Shan'           => [6378388.0,   1/297.0,         -634.0,  -549.0,  -201.0 ],
    'Indian'                => [6377276.645, 1/300.8017,      +289.0,  +734.0,  +257.0 ],
    'North American 1927'   => [6378206.4,   1/294.9786982,     -8.0,  +160.0,  +176.0 ],
    'North American 1983'   => [6378137.0,   1/298.257222101,    0.0,     0.0,     0.0 ],
    'Oman'                  => [6378249.145, 1/293.465,       -346.0,    -1.0,  +224.0 ],
    'Ordnance Survey 1936'  => [6377563.396, 1/299.3249646,   +375.0,  -111.0,  +431.0 ],
    'Pulkovo 1942'          => [6378245.0,   1/298.3,          +27.0,  -135.0,   -89.0 ],
    'Prov. S American 1956' => [6378388.0,   1/297.0,         -288.0,  +175.0,  -376.0 ],
    'South American 1969'   => [6378160.0,   1/298.25,         -57.0,    +1.0,   +41.0 ],
    'WGS72'                 => [6378135.0,   1/298.26,           0.0,     0.0,    -4.5 ]
  }

# 固定値パラメータを設定
  def initialize
    @rad = PI/180
  end

# 測地系パラメータを設定
  def setDatum(datum)
    @a, @f, @dx, @dy, @dz = Parameter[datum]
    @ee = (2 - @f) * @f  # 離心率の2乗
  end

# 楕円体座標を直交座標に変換 (lat:緯度, lon:経度, h:楕円体高)
  def calcECEF(lat, lon, h)
    lat *= @rad
    lon *= @rad
    lat_sin, lat_cos = sin(lat), cos(lat)
    n = @a / sqrt(1 - lat_sin ** 2 * @ee)
    [lat_cos * cos(lon) * (n + h) + @dx,    # x
     lat_cos * sin(lon) * (n + h) + @dy,    # y
     lat_sin * (n * (1 - @ee) + h) + @dz ]  # z
  end

# 直交座標を楕円体座標に変換
  def calcBLH(x, y, z)
    ba = 1 - @f  # b/a
    aee = @a * @ee
    x -= @dx
    y -= @dy
    z -= @dz
    xy = sqrt(x ** 2 + y ** 2)
    lat = atan2(z, xy * ba)
    lat = atan2(z + sin(lat) ** 3 * aee / ba, xy - cos(lat) ** 3 * aee)
    lon = atan2(y, x)
    n = @a / sqrt(1 - sin(lat) ** 2 * @ee)
    [lat / @rad, lon / @rad, xy / cos(lat) - n]  # lat, lon, h
  end

# 2直交座標間の方向を算出 ((x0,y0,z0)から見て(x,y,z)の方向)
  def calcDir(x, y, z, x0, y0, z0)
    lat = atan2(z0, sqrt(x0 ** 2 + y0 ** 2))  # 地心緯度
    lon = atan2(y0, x0)                       # 地心経度
    lon_sin, lon_cos = sin(lon), cos(lon)
    x, y = lon_cos * x + lon_sin * y, lon_cos * y - lon_sin * x
    x = sin(lat) * x - cos(lat) * z
    180 - atan2(y, x) / @rad
  end

# 2直交座標間の距離を算出
  def calcDis(x, y, z, x0, y0, z0)
    sqrt((x - x0) ** 2 + (y - y0) ** 2 + (z - z0) ** 2)
  end

# 度分秒の単位文字列を設定
  def setUnit(d, m, s)
    @d, @m, @s = d, m, s
    @d_re, @m_re, @s_re = Regexp.escape(@d), Regexp.escape(@m), Regexp.escape(@s)
  end

# 文字列を緯度/経度に変換
  def parse(d)
    if d.is_a?(String)
      case d
      when /^\s*([+-]?\d+)#{@d_re}(\d+)#{@m_re}(\d+(?:\.\d+)?)#{@s_re}\s*$/
        d, m, s = $1.to_i, $2.to_i, $3.to_f
        d = (d.abs + (m + s / 60) / 60) * (d < 0 ? -1 : +1)
      when /^\s*([+-]?\d+(?:\.\d+)?)\s*$/
        d = $1.to_f
      end
    end
    d
  end

# 緯度/経度をDMS文字列に変換
  def to_s(d)
    dms = [d < 0 ? '-' : '+']
    d = d.abs
    dms << d.to_i
    d = (d - dms.last) * 60
    dms << d.to_i
    d = (d - dms.last) * 60
    dms << d
    "%s%d#{@d}%d#{@m}%.3f#{@s}" % dms
  end

end

class Datum < CalcDatum

  def setECEF(x, y, z)
    @ecef = [x, y, z]
    @blh = {}
  end

  def setBLH(datum, lat, lon, h)
    lat, lon = parse(lat), parse(lon)
    setDatum(datum)
    @ecef = calcECEF(lat, lon, h)
    @blh = {datum => [lat, lon, h]}
  end

  def getECEF
    @ecef
  end

  def getBLH(datum)
    if !(blh = @blh[datum])
      setDatum(datum)
      blh = calcBLH(*@ecef)
      @blh[datum] = blh
    end
    blh
  end

  def to_s(datum)
    blh = getBLH(datum)
    [super(blh[0]), super(blh[1]), '%+.6f' % [blh[2]]]
  end

  def calcDir(x, y, z)
    super(x, y, z, *@ecef)
  end

  def calcDis(x, y, z)
    super(x, y, z, *@ecef)
  end

end

if $0 == __FILE__

  require 'cgi'
  require 'open-uri'

  cgi = nil

  begin

    cgi, html = CGI.new, []
    pos, mark, last, time, agent = nil, nil, nil, nil, nil

# クッキーから位置情報取得
    cgi.cookies[CookieName].each do |value|
      case value
      when /^mark=([+-]?\d+(?:\.\d+)?),([+-]?\d+(?:\.\d+)?),([+-]?\d+(?:\.\d+)?)$/  # 目標位置の直交座標
        mark = Datum.new
        mark.setECEF($1.to_f, $2.to_f, $3.to_f)
      when /^last=([+-]?\d+(?:\.\d+)?),([+-]?\d+(?:\.\d+)?),([+-]?\d+(?:\.\d+)?)$/  # 前回アクセス位置の直交座標
        last = Datum.new
        last.setECEF($1.to_f, $2.to_f, $3.to_f)
      when /^time=(\d+)$/                                                           # 前回アクセスの時間
        time = Time.at($1.to_i)
      when /^agent=(.+)$/
        agent = $1
      end
    end

# フォームフィールドから位置情報取得
    form_k = nil
    if cgi.query_string and cgi.query_string.length > 0
      form_lat, form_lon, form_h, form_datum = nil, nil, nil, nil
      agent_guess = ['willcom', 'au', 'softbank', 'docomo']
      cgi.query_string.split('&').each do |query|
        case CGI.unescape(query)
        when /^k=(.+)$/
          form_k = $1
          agent_guess &= []
# 携ナビ位置情報
        when /^kadr=(.+)$/
          form = ['q=' + CGI.escape($1),
                  'key=' + GoogleMapsAPIKey,
                  'sensor=false',
                  'output=csv' ]
          open('http://maps.google.co.jp/maps/geo?' + form.join('&')) do |f|
            f.each do |line|
              if line =~ /^200,.*?,([+-]?\d+(?:\.\d+){0,3}),([+-]?\d+(?:\.\d+){0,3})$/
                form_lat, form_lon = $1, $2
              end
            end
          end
          agent_guess &= []
        when /^kpos=([+-]?\d+(?:\.\d+)?),([+-]?\d+(?:\.\d+)?),([+-]?\d+(?:\.\d+)?)$/   # 現在位置の直交座標
          x, y, z = $1.to_f, $2.to_f, $3.to_f
          pos = Datum.new
          pos.setECEF(x, y, z)
          agent_guess &= []
        when /^kmark=([+-]?\d+(?:\.\d+)?),([+-]?\d+(?:\.\d+)?),([+-]?\d+(?:\.\d+)?)$/  # 目標位置の直交座標
          mark = Datum.new
          mark.setECEF($1.to_f, $2.to_f, $3.to_f)
          agent_guess &= []
        when /^klast=([+-]?\d+(?:\.\d+)?),([+-]?\d+(?:\.\d+)?),([+-]?\d+(?:\.\d+)?)$/  # 前回アクセス位置直の交座標
          last = Datum.new
          last.setECEF($1.to_f, $2.to_f, $3.to_f)
          agent_guess &= []
        when /^ktime=(\d+)$/                                                           # 前回アクセスの時間
          time = Time.at($1.to_i)
          agent_guess &= []
        when /^klat=([+-]?\d+(?:\.\d+){0,3})$/                                         # 現在位置の緯度
          form_lat = $1
          agent_guess &= []
        when /^klon=([+-]?\d+(?:\.\d+){0,3})$/                                         # 現在位置の経度
          form_lon = $1
          agent_guess &= []
        when /^kh=([+-]?\d+(?:\.\d+)?)$/                                               # 現在位置の楕円体高
          form_h = $1.to_f
          agent_guess &= []
        when /^kdatum=(.+)$/                                                           # 現在位置の測地系
          form_datum = $1
          agent_guess &= []
# 現在位置情報取得
        when /^pos=N999\.99\.99\.999E999\.99\.99\.999$/  # 取得失敗
          agent_guess &= []
        when /^pos=N(\d+\.\d+\.\d+\.\d+)E(\d+\.\d+\.\d+\.\d+)$/
          form_lat, form_lon = '+' + $1, '+' + $2
          form_datum = 'Tokyo'
          agent_guess &= ['willcom', 'softbank']
        when /^lat=([+-]?\d+\.\d+(?:\.\d+\.\d+)?)$/
          form_lat = $1
          agent_guess &= ['au', 'docomo']
        when /^lon=([+-]?\d+\.\d+(?:\.\d+\.\d+)?)$/
          form_lon = $1
          agent_guess &= ['au', 'docomo']
        when /^datum=(?:0|wgs84|WGS84)$/
          form_datum = 'WGS84'
          agent_guess &= ['au']
        when /^datum=(?:1|tokyo|TOKYO)$/
          form_datum = 'Tokyo'
          agent_guess &= ['au']
        when /^datum=(?:2|itrf|ITRF)$/  # 未対応
          form_datum = 'WGS84'
          agent_guess &= ['au']
        when /^geo=(?:wgs84|WGS84)$/
          form_datum = 'WGS84'
          agent_guess &= ['softbank', 'docomo']
        when /^geo=(?:tokyo|TOKYO)$/
          form_datum = 'Tokyo'
          agent_guess &= ['softbank', 'docomo']
        when /^geo=(?:itrf|ITRF)$/      # 未対応
          form_datum = 'WGS84'
          agent_guess &= ['softbank', 'docomo']
        end
      end
      if !pos and form_lat and form_lon
        pos = Datum.new
        pos.setUnit('.', '.', '')
        pos.setBLH(form_datum ? form_datum : 'WGS84', form_lat, form_lon, form_h ? form_h : 0)
        agent = agent_guess.first if agent_guess.size > 0
      end
    else
      form_k = 'nop'
    end

    cookie_expires = nil
# HTMLヘッダー生成
    html << '<html>'
    html << '<head>'
    html << '<meta http-equiv="Content-Type" content="text/html; charset=utf8">'
    html << '<title>%s</title>' % [HTMLTitle]
    html << '</head>'
    html << '<body>'
    if HeaderMessage
      html << '<div align="left">'
      html << '<small>%s</small>' % [HeaderMessage]
      html << '</div>'
      html << '<hr>'
    end
    html << '<div align="center">'
    if pos
# 目標登録URLへ変換
      form = ['kpos='  + CGI.escape('%+.6f,%+.6f,%+.6f' % pos.getECEF),
              'kmark=' + CGI.escape('%+.6f,%+.6f,%+.6f' % pos.getECEF) ]
      if last
        form << 'klast=' + CGI.escape('%+.6f,%+.6f,%+.6f' % last.getECEF)
        form << 'ktime=' + CGI.escape('%d' % [time.to_i]) if time
      end
      url_setmark = CGIURL + '?' + form.join('&amp;')
# 現在位置の地名を取得
      address = nil
      form = ['key=' + GoogleMapsAPIKey,
              'sensor=false',
              'output=csv',
              'll=' + CGI.escape('%.6f,%.6f' % pos.getBLH('WGS84')) ]
      open('http://maps.google.co.jp/maps/geo?' + form.join('&')) do |f|
        f.each do |line|
          address = $1 if line =~ /^200,.*?,"(?:日本,\s*)?(?:〒\d+(?:-\d+)?\s*)?(.+?)"/
        end
      end
# GoogleMap地図表示URLへ変換
      marker = []
      marker << '%.6f,%.6f,midreds' % pos.getBLH('WGS84')
      marker << '%.6f,%.6f,midredg' % mark.getBLH('WGS84') if mark
      path = ['rgb:0x0000ff,weight:5']
      path << '%.6f,%.6f' % last.getBLH('WGS84') if last
      path << '%.6f,%.6f' % pos.getBLH('WGS84')
      path << '%.6f,%.6f' % mark.getBLH('WGS84') if mark
      form = ['size=219x304',
              'maptype=mobile',
              'key=' + GoogleMapsAPIKey,
              'sensor=false' ]
      form << 'markers=' + CGI.escape(marker.join('|')) if marker.size >= 1
      form << 'path=' + CGI.escape(path.join('|')) if path.size >= 3
      url_mapwide = 'http://maps.google.co.jp/staticmap?' + form.join('&amp;')
      form << 'center=' + CGI.escape('%.6f,%.6f' % pos.getBLH('WGS84'))
      form << 'zoom=17'
      url_map = 'http://maps.google.co.jp/staticmap?' + form.join('&amp;')
# HTML位置情報生成
      if address
        html << '<small>'
        html << address
        html << '</small>'
        html << '<br>'
      end
      pos.setUnit('°', '&apos;', '&quot;')
      lat, lon = pos.to_s('WGS84')
      case lat
      when /^\+/
        lat = ['北緯', $']
      when /^-/
        lat = ['南緯', $']
      else
        lat = ['緯度', lat]
      end
      case lon
      when /^\+/
        lon = ['東経', $']
      when /^-/
        lon = ['西経', $']
      else
        lon = ['経度', lon]
      end
      html << '%s: %s' % lat
      html << '<br>'
      html << '%s: %s' % lon
      html << '<br>'
      html << '<small>'
      if last or mark
        html << '<a href="%s" accesskey="1">[1]地図表示</a>(<a href="%s" accesskey="2">[2]詳細</a>)' % [url_mapwide, url_map]
      else
        html << '<a href="%s" accesskey="1">[1]地図表示</a>' % [url_map]
      end
      html << '<a href="%s" accesskey="0">[0]目標に設定</a>' % [url_setmark]
      html << '</small>'
      if mark
        html << '<hr width="25%">'
        dis = pos.calcDis(*mark.getECEF)
        if dis > 0
          dir = pos.calcDir(*mark.getECEF)
          html << '目標への方向: %.1f° <img src="%s" width="16" height="16">' % [dir, ImageURL % [(dir / 10).round % 36]]
          html << '<br>'
        end
        html << '目標への距離: %.3fkm' % [dis * 1e-3]
        if last and time and (t = TimeNow - time) > 0
          if (d = last.calcDis(*mark.getECEF) - dis) > 0
            html << '<br>'
            h = dis / (d / t) / 3600
            hms = [h.to_i]
            h = (h - hms.last) * 60
            hms << h.to_i
            h = (h - hms.last) * 60
            hms << h
            html << '目標への時間: %d時間%d分' % hms
          end
        end
      end
      if last
        html << '<hr width="25%">'
        dis = pos.calcDis(*last.getECEF)
        if dis > 0
          dir = pos.calcDir(*last.getECEF)
          dir += dir < 180 ? +180 : -180
          html << '移動した方向: %.1f° <img src="%s" width="16" height="16">' % [dir, ImageURL % [(dir / 10).round % 36]]
          html << '<br>'
        end
        html << '移動した距離: %.3fkm' % [dis * 1e-3]
        if time and (t = TimeNow - time) > 0
          html << '<br>'
          html << '移動した速度: %.1fkm/h' % [dis / t * 3.6]
        end
      end
      cookie_expires = CookieExpires
    else
      case form_k
      when 'pos'
        html << '<div align="center">'
        html << '<form method="get" action="%s">' % [CGIURL]
        html << '緯度:'
        html << '<input type="text" name="klat" size="14" istyle="4">'
        html << '<br>'
        html << '経度:'
        html << '<input type="text" name="klon" size="14" istyle="4">'
        html << '<br><br>'
        html << '測地系:'
        html << '<input type="radio" name="kdatum" value="WGS84" checked>WGS84'
        html << '<input type="radio" name="kdatum" value="Tokyo">日本'
        html << '<br><br>'
        html << '<input type="submit" value="現在位置に設定">'
        html << '</form>'
        html << '</div>'
        cookie_expires = CookieExpires
      when 'adr'
        html << '<div align="center">'
        html << '<form method="get" action="%s">' % [CGIURL]
        html << '地名:'
        html << '<input type="text" name="kadr" size="14">'
        html << '<br><br>'
        html << '<input type="submit" value="現在位置に設定">'
        html << '</form>'
        html << '</div>'
        cookie_expires = CookieExpires
      when 'reset'
        html << '<p>クッキーから携ナビ情報を消去しました。</p>'
      else
        if form_k != 'nop'
          html << '位置情報取得失敗'
          html << '<hr width="25%">'
        end
        case agent
        when 'willcom'
          html << '<a href="http://location.request/dummy.cgi?my=%s&amp;pos=$location">位置情報送信</a>' % [CGIURL]
        when 'au'
          html << '<a href="device:gpsone?url=%s&amp;ver=1&amp;datum=0&amp;unit=0">位置情報送信(GPS対応機)</a>' % [CGIURL]
          html << '<br>'
          html << '<a href="device:location?url=%s">位置情報送信(GPS非対応機)</a>' % [CGIURL]
        when 'softbank'
          html << '<a href="location:auto?%s">位置情報送信</a>' % [CGIURL]
        when 'docomo'
          html << '<a href="%s" lcs>位置情報送信</a>' % [CGIURL]
        else
          html << '位置情報送信'
          html << '<br>'
          html << '<a href="http://location.request/dummy.cgi?my=%s&amp;pos=$location">WILLCOM</a>' % [CGIURL]
          html << '<br>'
          html << '<a href="device:gpsone?url=%s&amp;ver=1&amp;datum=0&amp;unit=0">au(GPS対応機)</a>' % [CGIURL]
          html << '<br>'
          html << '<a href="device:location?url=%s">au(GPS非対応機)</a>' % [CGIURL]
          html << '<br>'
          html << '<a href="location:auto?%s">SoftBank</a>' % [CGIURL]
          html << '<br>'
          html << '<a href="%s" lcs>DoCoMo</a>' % [CGIURL]
        end
        cookie_expires = CookieExpires
      end
    end
    html << '</div>'
# HTMLフッダー生成
    html << '<hr>'
    if FooterMessage
      html << '<div align="left">'
      html << '<small>%s</small>' % [FooterMessage]
      html << '</div>'
    end
    html << '<div align="right">'
    html << '<small><a href="%s?k=adr" accesskey="*">[*]地名入力</a></small>' % [CGIURL]
    html << '<small><a href="%s?k=pos" accesskey="#">[#]座標入力</a></small>' % [CGIURL]
    html << '<small><a href="%s">Help</a></small>' % [HelpURL]
    html << '<small><a href="%s">BBS</a></small>' % [BBSURL] if BBSURL
    html << '</div>'
    html << '</body>'
    html << '</html>'
    html = html.join("\n") + "\n"

# 目標位置と現在位置の直交座標をクッキーに残す
    cookie_value = []
    cookie_value << 'mark=%+.6f,%+.6f,%+.6f' % mark.getECEF if mark
    if pos
      cookie_value << 'last=%+.6f,%+.6f,%+.6f' % pos.getECEF
      cookie_value << 'time=%d' % [TimeNow.to_i]
    elsif last
      cookie_value << 'last=%+.6f,%+.6f,%+.6f' % last.getECEF
      cookie_value << 'time=%d' % [time.to_i] if time
    end
    cookie_value << 'agent=%s' % [agent] if agent
    if cookie_value.size > 0
      cookie = CGI::Cookie.new
      cookie.name, cookie.value = CookieName, cookie_value
      cookie.expires = CookieExpires if cookie_expires
      cgi.out('cookie' => [cookie]) {html}
    else
      cgi.out {html}
    end

  rescue
    cgi.out('Content-Type' => 'text/plain') {$!.to_s} if cgi

  end

end
